from enum import Enum

try:
    from typing import Literal, Union
except ImportError:
    from typing_extensions import Literal
from typing import Optional

_BASE_OPTION_SETS = [
	#no jailbreak filter
    "nojbf",
    "iyxapbing",
    "iycapbing",
    "dgencontentv3",
    "nointernalsugg",
    "disable_telemetry",
    "machine_affinity",
    "streamf",
    "langdtwb",
    "fdwtlst",
    # "fluxcrplus",
    "fluxprod"
    "eredirecturl",
    # may related to image search
    "gptvnodesc",
    "gptvnoex",
    "sdretrieval",
    "gamaxinvoc",
    "ldsummary",
    "ldqa",
]

class ConversationStyle(Enum):
    creative = _BASE_OPTION_SETS + ["Creative"]
    balanced = _BASE_OPTION_SETS + ["galileo", "gldcl1p"]
    precise = _BASE_OPTION_SETS + ["h3precise"]


CONVERSATION_STYLE_TYPE = Optional[
    Union[ConversationStyle, Literal["creative", "balanced", "precise"]]
]

class Persona(Enum):
    designer = "ai_persona_designer_gpt"
    travel = "flux_vacation_planning_helper_v14"
    cooking = "flux_cooking_helper_v14"
    fitness = "flux_fitness_helper_v14"
    copilot = "fluxcopilot"
    sydney = "fluxsydney"



from enum import Enum

try:
    from typing import Literal, Union
except ImportError:
    from typing_extensions import Literal
from typing import Optional


class LocationHint(Enum):
    USA = {
        "locale": "en-US",
        "LocationHint": [
            {
                "SourceType": 1,
                "RegionType": 2,
                "Center": {
                    "Latitude": 34.05189895529883,
                    "Longitude": -118.26219950185547
                },
                "Radius": 24902,
                "Name": "Los Angeles, California",
                "Accuracy": 24902,
                "FDConfidence": 0.5,
                "CountryName": "United States",
                "CountryConfidence": 8,
                "Admin1Name": "California",
                "PopulatedPlaceName": "Los Angeles",
                "PopulatedPlaceConfidence": 5,
                "PostCodeName": "90017",
                "UtcOffset": -8,
                "Dma": 803
            }
        ]
    }
    CHINA = {
        "locale": "zh-CN",
        "LocationHint": [
            {
                "SourceType": 1,
                "RegionType": 2,
                "Center": {
                    "Latitude": 39.9042,
                    "Longitude": 116.4074,
                },
                "Radius": 24902,
                "Name": "Beijing",
                "Accuracy": 24902,
                "FDConfidence": 0.5,
                "CountryName": "China",
                "CountryConfidence": 8,
                "Admin1Name": "Kowloon City",
                "PopulatedPlaceName": "Kowloon City District",
                "PopulatedPlaceConfidence": 5,
                "UtcOffset": 8,
                "Dma": 0
            },
        ],
    }
    HONGKONG = {
        "locale": "zh-HK",
        "LocationHint": [
            {
                "SourceType": 1,
                "RegionType": 2,
                "Center": {
                    "Latitude": 22.32819938659668,
                    "Longitude": 114.19159698486328
                },
                "Radius": 24902,
                "Name": "Kowloon City District, Kowloon City",
                "Accuracy": 24902,
                "FDConfidence": 0.5,
                "CountryName": "Hong Kong",
                "CountryConfidence": 8,
                "Admin1Name": "Kowloon City",
                "PopulatedPlaceName": "Kowloon City District",
                "PopulatedPlaceConfidence": 5,
                "UtcOffset": 8,
                "Dma": 0
            },
        ],
    }
    TW = {
        "locale": "zh-TW",
        "LocationHint": [
            {
                "SourceType": 1,
                "RegionType": 2,
                "Center": {
                    "Latitude": 24.14189910888672,
                    "Longitude": 120.68060302734375
                },
                "Radius": 24902,
                "Name": "Central District, Taichung City",
                "Accuracy": 24902,
                "FDConfidence": 0.800000011920929,
                "CountryName": "Taiwan",
                "CountryConfidence": 8,
                "Admin1Name": "Taichung City",
                "PopulatedPlaceName": "Central District",
                "PopulatedPlaceConfidence": 8,
                "UtcOffset": 8,
                "Dma": 0
            },
        ],
    }
    EU = {
        "locale": "en-IE",
        "LocationHint": [
            {
                "country": "Norway",
                "state": "",
                "Name": "Oslo",
                "timezoneoffset": 1,
                "countryConfidence": 8,
                "Center": {
                    "Latitude": 59.9139,
                    "Longitude": 10.7522,
                },
                "RegionType": 2,
                "SourceType": 1,
            },
        ],
    }
    UK = {
        "locale": "en-GB",
        "LocationHint": [
            {
                "country": "United Kingdom",
                "state": "",
                "Name": "London",
                "timezoneoffset": 0,
                "countryConfidence": 8,
                "Center": {
                    "Latitude": 51.5074,
                    "Longitude": -0.1278,
                },
                "RegionType": 2,
                "SourceType": 1,
            },
        ],
    }


LOCATION_HINT_TYPES = Optional[Union[LocationHint, Literal["USA", "CHINA", "HONGKONG","EU", "UK"]]]



import asyncio
import base64
import json
import os
from time import time
from typing import Generator, List, Union, Optional, Tuple

from curl_cffi import requests
from curl_cffi.requests.websockets import WebSocket
from curl_cffi.const import CurlWsFlag
from curl_cffi import AsyncCurl
from curl_cffi.curl import CurlMime
import urllib
from BingImageCreator import ImageGenAsync

from .constants import DELIMITER, PERSONATE
from .constants import HEADERS
from .constants import HEADERS_INIT_CONVER
from .conversation import Conversation
from .conversation_style import CONVERSATION_STYLE_TYPE, Persona
from .plugin import Plugin
from .request import ChatHubRequest
from .utilities import append_identifier
from .utilities import get_ran_hex
from .utilities import guess_locale
from .utilities import parse_search_result


class ChatHub:
    def __init__(
            self,
            conversation: Conversation,
            proxy: str = None,
            cookies: Union[List[dict], None] = None,
    ) -> None:
        conversationSignature = conversation.struct.get("conversationSignature") or None
        self.aio_session = None
        self.request: ChatHubRequest
        self.loop: bool
        self.task: asyncio.Task
        self.request = ChatHubRequest(
            conversation_signature=conversationSignature,
            client_id=conversation.struct["clientId"],
            conversation_id=conversation.struct["conversationId"],
        )
        self.sec_access_token = conversation.struct["secAccessToken"] or None
        self.cookies = cookies
        self.proxy: str = proxy
        proxy = (
                proxy
                or os.environ.get("all_proxy")
                or os.environ.get("ALL_PROXY")
                or os.environ.get("https_proxy")
                or os.environ.get("HTTPS_PROXY")
                or None
        )
        if proxy is not None and proxy.startswith("socks5h://"):
            proxy = "socks5://" + proxy[len("socks5h://"):]
        self.session = AsyncSession(
            proxy=proxy,
            timeout=900,
            headers=HEADERS_INIT_CONVER,
            impersonate=PERSONATE
        )

    async def get_conversation(
            self,
            conversation_id: str = None,
            conversation_signature: str = None,
            client_id: str = None,
    ) -> dict:
        conversation_id = conversation_id or self.request.conversation_id
        conversation_signature = (
                conversation_signature or self.request.conversation_signature
        )
        client_id = client_id or self.request.client_id
        url = f"https://sydney.bing.com/sydney/GetConversation?conversationId={conversation_id}&source=cib&participantId={client_id}&conversationSignature={conversation_signature}&traceId={get_ran_hex()}"
        response = await self.session.get(url)
        return response.json()

    async def get_activity(self) -> dict:
        url = "https://www.bing.com/turing/conversation/chats"
        headers = HEADERS_INIT_CONVER.copy()
        if self.cookies is not None:
            for cookie in self.cookies:
                if cookie["name"] == "_U":
                    headers["Cookie"] = f"SUID=A; _U={cookie['value']};"
                    break
        response = await self.session.get(url, headers=headers)
        return response.json()

    async def ask_stream(
            self,
            prompt: str,
            wss_link: str = None,
            conversation_style: CONVERSATION_STYLE_TYPE = None,
            raw: bool = False,
            webpage_context: Union[str, None] = None,
            locale: str = guess_locale(),
            no_search: bool = True,
            persona: Persona = Persona.copilot,
            plugins: set[Plugin] = {},
            image_url: str = None,
    ) -> Generator[bool, Union[dict, str], None]:
        """ """
        if (not prompt) and (not image_url):
            raise ValueError("prompt or image_url must be provided")

        cookies = {}
        if self.cookies is not None:
            for cookie in self.cookies:
                cookies[cookie["name"]] = cookie["value"]
        # self.aio_session = aiohttp.ClientSession(cookies=cookies)
        self.aio_session = AsyncSession(
            cookies=cookies,
            headers=HEADERS,
            proxy=self.proxy,
            impersonate=PERSONATE
        )
        token = urllib.parse.quote_plus(self.sec_access_token) if self.sec_access_token else ''
        wss_link = f'{wss_link}?sec_access_token={token}'
        # Check if websocket is closed
        wss = await self.aio_session.ws_connect(wss_link or "wss://sydney.bing.com/sydney/ChatHub")
        await self._initial_handshake(wss)
        # Construct a ChatHub request
        self.request.update(
            prompt=prompt,
            conversation_style=conversation_style,
            webpage_context=webpage_context,
            locale=locale,
            no_search=no_search,
            persona=persona,
            plugins=plugins,
            image_url=image_url,
        )
        # Send request
        await wss.asend(append_identifier(self.request.struct).encode("utf-8"))
        resp_txt = ""
        generate = None
        search_refs = []
        search_keywords = []
        async for obj in self._receive_messages(wss):
            if int(time()) % 6 == 0:
                await wss.asend(append_identifier({"type": 6}).encode("utf-8"))

            if obj is None or not obj:
                continue

            response = json.loads(obj)
            r_type = response.get("type")
            if r_type in [6, 7]:
                await wss.asend(append_identifier({"type": r_type}).encode("utf-8"))

            if raw:
                done = r_type == 2
                yield done, response
                if not done:
                    continue
                else:
                    break

            if r_type == 1 and response["arguments"][0].get(
                    "messages",
            ):
                message = response["arguments"][0]["messages"][0]
                msg_type = message.get("messageType")
                hidden_text = message.get("hiddenText")
                text = message.get("text")
                if msg_type == "InternalSearchQuery":
                    search_keywords.append(hidden_text)
                elif msg_type == "InternalSearchResult":
                    search_refs += parse_search_result(message)
                elif msg_type == "GenerateContentQuery":
                    generate = {
                        "content_type": message.get("contentType"),
                        "prompt": text
                    }
                elif msg_type is None:
                    if message.get("contentOrigin") == "Apology":
                        print('message has been revoked')
                        print(message)
                        resp_txt = f"{resp_txt}{text} -end- (message has been revoked)"

                    if len(search_keywords) > 0:
                        keywords = "\n* ".join(search_keywords)
                        resp_txt = f"{resp_txt}Searching the web for:\n* {keywords}\n\n"
                        search_keywords = []

                    yield False, f"{resp_txt}{text}"
            elif response.get("type") == 2:
                if response["item"]["result"].get("error"):
                    raise Exception(
                        f"{response['item']['result']['value']}: {response['item']['result']['message']}",
                    )

                response["media"] = {}
                message = response["item"]["messages"][-1]
                text = message.get("text")
                if generate:
                    if generate["content_type"] == "IMAGE":
                        async with ImageGenAsync(
                                all_cookies=self.cookies,
                                proxy=self.proxy
                        ) as image_obj:
                            try:
                                images = await image_obj.get_images(generate["prompt"])
                                response["media"] = {
                                    "prompt": generate["prompt"],
                                    "images": images
                                }
                            except Exception as e:
                                print(str(e))
                                hint = "Your prompt has been prohibited by third-service. Please modify it."
                                resp_txt = f"{resp_txt}{text}\n{e}\n{hint}"

                if len(search_refs) > 0:
                    refs_str = ""
                    for index, item in enumerate(search_refs):
                        refs_str += f'- [^{index}^] [{item["title"]}]({item["url"]})\n'

                    resp_txt = f"{resp_txt}{text}\n{refs_str}"

                message["text"] = resp_txt
                if "media" not in response:
                    response["media"] = {}

                yield True, response
                return

    async def _initial_handshake(self, wss) -> None:
        proto = append_identifier({"protocol": "json", "version": 1})
        await wss.asend(proto.encode("utf-8"))
        await wss.arecv()
        await wss.asend(append_identifier({"type": 6}).encode("utf8"))

    async def _receive_messages(self, wss: WebSocket) -> Generator[str, Tuple[bytes, int], None]:
        buffer = b''
        retry_count = 5
        is_connected = True
        while is_connected:
            # Receive the next chunk of data
            chunk, flags = await wss.arecv()

            if flags & CurlWsFlag.CLOSE:
                is_connected = False

            if not chunk:
                retry_count -= 1
                if retry_count == 0:
                    raise Exception("No response from server")
                continue

            buffer += chunk
            # Split the buffer by the delimiter
            while b'\x1e' in buffer:
                delimiter_index = buffer.index(b'\x1e')
                # Extract the complete message up to the delimiter
                complete_message = buffer[:delimiter_index]
                yield complete_message.decode('utf-8')
                # Remove the processed message from the buffer
                buffer = buffer[delimiter_index + 1:]

        return

    async def delete_conversation(
            self,
            conversation_id: str = None,
            conversation_signature: str = None,
            client_id: str = None,
    ) -> None:
        conversation_id = conversation_id or self.conversation_id
        conversation_signature = (
                conversation_signature or self.request.conversation_signature
        )
        client_id = client_id or self.request.client_id
        url = "https://sydney.bing.com/sydney/DeleteSingleConversation"
        await self.session.post(
            url,
            json={
                "conversationId": conversation_id,
                "conversationSignature": conversation_signature,
                "participant": {"id": client_id},
                "source": "cib",
                "optionsSets": ["autosave"],
            },
        )

    async def close(self) -> None:
        await self.session.close()

    async def upload_image(
            self,
            binary_image: bytes,
            style: CONVERSATION_STYLE_TYPE = "precise",
            blur: bool = False
    ) -> dict:
        if not isinstance(style, str):
            style = style.name

        url = "https://www.bing.com/images/kblob"
        payload = {
            "imageInfo": {},
            "knowledgeRequest": {
                "invokedSkills": ["ImageById"],
                "subscriptionId": "Bing.Chat.Multimodal",
                "invokedSkillsRequestData": {"enableFaceBlur": blur},
                "convoData": {
                    "convoid": "",
                    "convotone": style,
                }
            }
        }
        encode_data = base64.b64encode(binary_image)

        parts = CurlMime()
        parts.addpart(
            name="knowledgeRequest",
            content_type="application/json",
            data=json.dumps(payload).encode("utf-8"),
        )
        parts.addpart(
            name="imageBase64",
            content_type="application/octet-stream",
            data=encode_data,
        )

        async with AsyncSession(
            headers={
                "Referer": "https://www.bing.com/search?q=Bing+AI&showconv=1&FORM=hpcodx"
            },
            proxy=self.proxy,
            impersonate=PERSONATE,
            timeout=900,
        ) as session:
            resp = await session.post(url, multipart=parts)
            if not resp.ok:
                raise Exception(f"Failed to upload image. Status: {resp.status}")

            return resp.json()


class AsyncSession(requests.AsyncSession):
    def __init__(
            self,
            *,
            loop=None,
            async_curl: Optional[AsyncCurl] = None,
            max_clients: int = 10,
            **kwargs,
    ):
        super().__init__(loop=loop, async_curl=async_curl, max_clients=max_clients, **kwargs)

    def is_closed(self):
        return self._closed

    async def close(self) -> None:
        if not self._closed:
            await super().close()



import asyncio
import contextlib
import json
import re
from pathlib import Path

from log2d import Log

from .EdgeGPT import Chatbot
from .EdgeGPT import ConversationStyle
from .ImageGen import ImageGen

Log("BingChat")
log = Log.BingChat.debug  # shortcut to create a log entry


class Cookie:
    """
    Convenience class for Bing Cookie files, data, and configuration. This Class
    is updated dynamically by the Query class to allow cycling through >1
    cookie/credentials file e.g. when daily request limits (current 200 per
    account per day) are exceeded.
    """

    current_file_index = 0
    dir_path = Path.home().resolve() / "bing_cookies"
    current_file_path = dir_path  # Avoid Path errors when no cookie file used
    search_pattern = "bing_cookies_*.json"
    ignore_files = set()
    request_count = {}
    supplied_files = set()
    rotate_cookies = True

    @classmethod
    def files(cls: type) -> list[Path]:
        """
        Return a sorted list of all cookie files matching .search_pattern in
        cls.dir_path, plus any supplied files, minus any ignored files.
        """
        all_files = set(Path(cls.dir_path).glob(cls.search_pattern))
        if hasattr(cls, "supplied_files"):
            supplied_files = {x for x in cls.supplied_files if x.is_file()}
            all_files.update(supplied_files)
        return sorted(all_files - cls.ignore_files)

    @classmethod
    def import_data(cls: type) -> None:
        """
        Read the active cookie file and populate the following attributes:

          .current_file_path
          .current_data
          .image_token
        """
        if not cls.files():
            log("No files in Cookie.dir_path")
            return
        try:
            cls.current_file_path = cls.files()[cls.current_file_index]
        except IndexError:
            log(f"Invalid file index [{cls.current_file_index}]")
            log("Files in Cookie.dir_path:")
            for file in cls.files():
                log(f"{file}")
            return
        log(f"Importing cookies from: {cls.current_file_path.name}")
        with open(cls.current_file_path, encoding="utf-8") as file:
            cls.current_data = json.load(file)
        cls.image_token = [
            x for x in cls.current_data if x.get("name").startswith("_U")
        ]
        cls.image_token = cls.image_token[0].get("value")

    @classmethod
    def import_next(cls: type, discard: bool = False) -> None:
        """
        Cycle through to the next cookies file then import it.

        discard (bool): True -Mark the previous file to be ignored for the remainder of the current session.  Otherwise cycle through all available
        cookie files (sharing the workload and 'resting' when not in use).
        """
        if not hasattr(cls, "current_file_path"):
            cls.import_data()
            return
        with contextlib.suppress(AttributeError):
            # Will fail on first instantiation because no current_file_path
            if discard:
                cls.ignore_files.add(cls.current_file_path)
            else:
                Cookie.current_file_index += 1
        if Cookie.current_file_index >= len(cls.files()):
            Cookie.current_file_index = 0
        Cookie.import_data()


class Query:
    """
    A convenience class that wraps around EdgeGPT.Chatbot to encapsulate input,
    config, and output all together.  Relies on Cookie class for authentication
    unless ignore_cookies=True
    """

    index = []
    image_dir_path = Path.cwd().resolve() / "bing_images"

    def __init__(
        self,
        prompt: str,
        style: ConversationStyle = "precise",
        content_type: str = "text",
        cookie_files: set[Path] = None,
        ignore_cookies: bool = False,
        echo: bool = True,
        echo_prompt: bool = False,
        locale: str = "en-GB",
        simplify_response: bool = True,
    ) -> None:
        """
        Arguments:

        prompt: Text to enter into Bing Chat
        style: creative, balanced, or precise
        content_type: "text" for Bing Chat; "image" for Dall-e
        ignore_cookies (bool): Ignore cookie data altogether
        echo: Print something to confirm request made
        echo_prompt: Print confirmation of the evaluated prompt
        simplify_response: True -> single simplified prompt/response exchange
        cookie_files: iterable of Paths or strings of cookie files (json)

        Files in Cookie.dir_path will also be used if they exist.  This defaults
        to the current working directory, so set Cookie.dir_path before
        creating a Query if your cookie files are elsewhere.
        """
        self.__class__.index += [self]
        self.prompt = prompt
        self.locale = locale
        self.simplify_response = simplify_response
        self.ignore_cookies = ignore_cookies
        if not ignore_cookies:
            if cookie_files:
                # Convert singular argument to an iterable:
                if isinstance(cookie_files, (str, Path)):
                    cookie_files = {cookie_files}
                # Check all elements exist and are Paths:
                cookie_files = {
                    Path(x).resolve()
                    for x in cookie_files
                    if isinstance(x, (str, Path)) and x
                }
                Cookie.supplied_files = cookie_files
            files = Cookie.files()  # includes .supplied_files
            if Cookie.rotate_cookies:
                Cookie.import_next()
            else:
                Cookie.import_data()
        if content_type == "text":
            self.style = style
            self.log_and_send_query(echo, echo_prompt)
        if content_type == "image":
            self.create_image()

    def log_and_send_query(self, echo: bool, echo_prompt: bool) -> None:
        self.response = asyncio.run(self.send_to_bing(echo, echo_prompt))
        if not hasattr(Cookie, "current_data"):
            name = "<no_cookies>"
        else:
            name = Cookie.current_file_path.name
        if not Cookie.request_count.get(name):
            Cookie.request_count[name] = 1
        else:
            Cookie.request_count[name] += 1

    def create_image(self) -> None:
        image_generator = ImageGen(Cookie.image_token)
        image_generator.save_images(
            image_generator.get_images(self.prompt),
            output_dir=self.__class__.image_dir_path,
        )

    async def send_to_bing(self, echo: bool = True, echo_prompt: bool = False) -> str:
        """Creat, submit, then close a Chatbot instance.  Return the response"""
        retries = len(Cookie.files()) or 1
        while retries:
            if not hasattr(Cookie, "current_data"):
                bot = await Chatbot.create()
            else:
                bot = await Chatbot.create(cookies=Cookie.current_data)
            if echo_prompt:
                log(f"{self.prompt=}")
            if echo:
                log("Waiting for response...")
            if self.style.lower() not in "creative balanced precise".split():
                self.style = "precise"
            try:
                return await bot.ask(
                    prompt=self.prompt,
                    conversation_style=getattr(ConversationStyle, self.style),
                    simplify_response=self.simplify_response,
                    locale=self.locale,
                )
            except Exception as ex:
                log(
                    f"Exception: [{Cookie.current_file_path.name} may have exceeded the daily limit]\n{ex}",
                )
                Cookie.import_next(discard=True)
                retries -= 1
            finally:
                await bot.close()
        return None

    @property
    def output(self) -> str:
        """The response from a completed Chatbot request"""
        if not self.simplify_response:
            return [
                x.get("text") or x.get("hiddenText")
                for x in self.response["item"]["messages"]
                if x["author"] == "bot"
            ]
        try:
            return self.response["text"]
        except TypeError as te:
            raise TypeError(
                f"{te}\n(No response received - probably rate throttled...)",
            ) from te

    @property
    def sources(self) -> list[list[dict]]:
        """The source names and details parsed from a completed Chatbot request"""
        if self.simplify_response:
            return self.response["sources_text"]
        return [
            x.get("sourceAttributions") or []
            for x in self.response["item"]["messages"]
            if x["author"] == "bot"
        ]

    @property
    def sources_dict(self) -> dict[int, str]:
        """The source names and details as a dictionary"""
        if self.simplify_response:
            text = self.response["sources_text"]
            sources = enumerate(re.findall(r"\((http.*?)\)", text))
            return {index + 1: value for index, value in sources}
        all_sources = []
        name = "providerDisplayName"
        url = "seeMoreUrl"
        for sources in self.sources:
            if not sources:
                continue
            data = {
                index + 1: source[url]
                for index, source in enumerate(sources)
                if name in source and url in source
            }
            all_sources += [data]
        return all_sources

    @property
    def code_block_formats(self) -> list[str]:
        """
        Extract a list of programming languages/formats used in code blocks
        """
        regex = r"``` *(\b\w+\b\+*) *"
        if self.simplify_response:
            return re.findall(regex, self.output)
        return re.findall(regex, "\n".join(self.output))

    @property
    def code_blocks(self) -> list[str]:
        """
        Return a list of code blocks (```) or snippets (`) as strings.

        If the response contains a mix of snippets and code blocks, return the
        code blocks only.

        This method is not suitable if the main text response includes either of
        the delimiters but not as part of an actual snippet or code block.

        For example:
        'In Markdown, the back-tick (`) is used to denote a code snippet'

        """

        final_blocks = []
        if isinstance(self.output, str):  # I.e. simplify_response is True
            separator = "```" if "```" in self.output else "`"
            code_blocks = self.output.split(separator)[1:-1:2]
            if separator == "`":
                return code_blocks
        else:
            code_blocks = []
            for response in self.output:
                separator = "```" if "```" in response else "`"
                code_blocks.extend(response.split(separator)[1:-1:2])
            code_blocks = [x for x in code_blocks if x]
        # Remove language name if present:
        for block in code_blocks:
            lines = block.splitlines()
            code = lines[1:] if re.match(" *\\w+ *", lines[0]) else lines
            final_blocks += ["\n".join(code).removeprefix(separator)]
        return [x for x in final_blocks if x]

    @property
    def code(self) -> str:
        """
        Extract and join any snippets of code or formatted data in the response
        """
        return "\n\n".join(self.code_blocks)

    @property
    def suggestions(self) -> list[str]:
        """Follow-on questions suggested by the Chatbot"""
        if self.simplify_response:
            return self.response["suggestions"]
        try:
            return [
                x["text"]
                for x in self.response["item"]["messages"][1]["suggestedResponses"]
            ]
        except KeyError:
            return None

    def __repr__(self) -> str:
        return f"<EdgeGPT.Query: {self.prompt}>"

    def __str__(self) -> str:
        return self.output if self.simplify_response else "\n\n".join(self.output)


class ImageQuery(Query):
    def __init__(self, prompt: str, **kwargs) -> None:
        kwargs["content_type"] = "image"
        super().__init__(prompt, **kwargs)

    def __repr__(self) -> str:
        return f"<EdgeGPT.ImageQuery: {self.prompt}>"


def test_cookie_rotation() -> None:
    for i in range(1, 50):
        q = Query(
            f"What is {i} in Roman numerals?  Give the answer in JSON",
            style="precise",
        )
        log(f"{i}: {Cookie.current_file_path.name}")
        log(q.code)
        log(f"Cookie count: {Cookie.request_count.get(Cookie.current_file_path.name)}")


def test_features() -> Query:
    try:
        q = Query(
            f"What is {i} in Roman numerals?  Give the answer in JSON",
            style="precise",
        )
        log(f"{i}: {Cookie.current_file_path.name}")
        print(f"{Cookie.current_file_index=}")
        print(f"{Cookie.current_file_path=}")
        print(f"{Cookie.current_data=}")
        print(f"{Cookie.dir_path=}")
        print(f"{Cookie.search_pattern=}")
        print(f"{Cookie.files()=}")
        print(f"{Cookie.image_token=}")
        print(f"{Cookie.import_next(discard=True)=}")
        print(f"{Cookie.rotate_cookies=}")
        print(f"{Cookie.files()=}")
        print(f"{Cookie.ignore_files=}")
        print(f"{Cookie.supplied_files=}")
        print(
            f"{Cookie.request_count=}"
        )  # Keeps a tally of requests made in using each cookie file during this session
        print(f"{q=}")
        print(f"{q.prompt=}")
        print(f"{q.ignore_cookies=}")
        print(f"{q.style=}")
        print(f"{q.simplify_response=}")
        print(f"{q.locale=}")
        print(f"{q.output=}")
        print(q)
        print(f"{q.sources=}")
        print(f"{q.sources_dict=}")
        print(f"{q.suggestions=}")
        print(f"{q.code=}")  # All code as a single string
        print(f"{q.code_blocks=}")  # Individual code blocks
        print(
            f"{q.code_block_formats=}"
        )  # The language/format of each code block (if given)
        print(f"{Query.index=}")  # Keeps an index of Query objects created
        print(f"{Query.image_dir_path=}")
    except Exception as E:
        raise Exception(E) from E
    finally:
        return q



from enum import Enum

class Item():
    def __init__(self,id: str, option_set: str) -> None:
        self.id = id
        self.option_set = option_set
    

class Plugin(Enum):
    codeInterpreter = Item(None,"codeintfile")
    instacart = Item('46664d33-1591-4ce8-b3fb-ba1022b66c11', '0A402EDC')
    kayak = Item('d6be744c-2bd9-432f-95b7-76e103946e34', 'C0BB4EAB')
    klarna = Item('5f143ea3-8c80-4efd-9515-185e83b7cf8a', '606E9E5D')
    openTable = Item('543a7b1b-ebc6-46f4-be76-00c202990a1b', 'E05D72DE')
    search = Item('c310c353-b9f0-4d76-ab0d-1dd5e979cf68', None)
    shop = Item('39e3566a-d481-4d99-82b2-6d739b1e716e', '2E842A93')
    suno = Item('22b7f79d-8ea4-437e-b5fd-3e21f09f7bc1', '014CB21D')



import json
import locale
import random
import sys
from typing import Union

from .constants import DELIMITER
from .locale import LocationHint


def append_identifier(msg: dict) -> str:
    # Convert dict to json string
    return json.dumps(msg, ensure_ascii=False) + DELIMITER


def get_ran_hex(length: int = 32) -> str:
    return "".join(random.choice("0123456789abcdef") for _ in range(length))


def get_location_hint_from_locale(locale: str) -> Union[dict, None]:
    locale = locale.lower()
    if locale == "en-gb":
        hint = LocationHint.UK.value
    elif locale == "en-ie":
        hint = LocationHint.EU.value
    elif locale == "zh-cn":
        hint = LocationHint.CHINA.value
    elif locale == "zh-hk":
        hint = LocationHint.HONGKONG.value
    elif locale == "zh-tw":
        hint = LocationHint.TW.value
    else:
        hint = LocationHint.USA.value
    return hint.get("LocationHint")


def guess_locale() -> str:
    if sys.platform.startswith("win"):
        return "en-us"
    loc, _ = locale.getlocale()
    return loc.replace("_", "-") if loc else "en-us"


def parse_search_result(message):
    if 'Web search returned no relevant result' in message['hiddenText']:
        return [{
            'title': 'No relevant result',
            'url': None,
            'snippet': message['hiddenText']
        }]

    data = []
    for group in json.loads(message['text']).values():
        for item in group:
            data.append({
                'title': item['title'],
                'url': item['url'],
            })

    return data



import uuid
from datetime import datetime
from typing import Union

from .conversation_style import CONVERSATION_STYLE_TYPE
from .conversation_style import ConversationStyle, Persona
from .plugin import Plugin
from .utilities import get_location_hint_from_locale
from .utilities import get_ran_hex
from .utilities import guess_locale
import json


class ChatHubRequest:
    def __init__(
            self,
            conversation_signature: str,
            client_id: str,
            conversation_id: str,
            invocation_id: int = 3,
    ) -> None:
        self.struct: dict = {}

        self.client_id: str = client_id
        self.conversation_id: str = conversation_id
        self.conversation_signature: str = conversation_signature
        self.invocation_id: int = invocation_id

    def update(
            self,
            prompt: str,
            conversation_style: CONVERSATION_STYLE_TYPE,
            webpage_context: Union[str, None] = None,
            locale: str = guess_locale(),
            no_search: bool = True,
            persona: Persona = Persona.copilot,
            plugins: set[Plugin] = {},
            image_url: str = None,
    ) -> None:
        options = [
            "deepleo",
            "enable_debug_commands",
            "disable_emoji_spoken_text",
            "enablemm",
        ]
        if conversation_style:
            if not isinstance(conversation_style, ConversationStyle):
                conversation_style = getattr(ConversationStyle, conversation_style)
            options = conversation_style.value.copy()

        plugin_params = []
        is_search_needed_for_plugin = False
        for plugin in plugins:
            val = plugin.value
            if val.id is not None:
                plugin_params.append({
                    "id": val.id,
                    "category": 1
                })

            if val.option_set is not None:
                options.append(val.option_set)

            if (not is_search_needed_for_plugin) and plugin != Plugin.codeInterpreter:
                is_search_needed_for_plugin = True

        if is_search_needed_for_plugin and (Plugin.search not in plugins):
            plugin_params.append({
                "id": Plugin.search.value.id,
                "category": 1
            })

        if (not is_search_needed_for_plugin) and no_search:
            options.append('nosearchall')

        options.append(persona.value)

        message_id = str(uuid.uuid4())
        # Get the current local time
        now_local = datetime.now()

        # Get the current UTC time
        now_utc = datetime.utcnow()

        # Calculate the time difference between local and UTC time
        timezone_offset = now_local - now_utc

        # Get the offset in hours and minutes
        offset_hours = int(timezone_offset.total_seconds() // 3600)
        offset_minutes = int((timezone_offset.total_seconds() % 3600) // 60)

        # Format the offset as a string
        offset_string = f"{offset_hours:+03d}:{offset_minutes:02d}"

        # Get current time
        timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%S") + offset_string
        message = {
            "locale": locale,
            "market": locale,
            "region": locale[-2:],  # en-US -> US
            "locationHints": get_location_hint_from_locale(locale),
            "timestamp": timestamp,
            "author": "user",
            "inputMethod": "Keyboard",
            "text": prompt,
            "messageType": "Chat",
            "messageId": message_id,
            "requestId": message_id,
        }
        if image_url is not None:
            message["imageUrl"] = image_url

        self.struct = {
            "arguments": [
                {
                    "source": "cib",
                    "optionsSets": options,
                    "allowedMessageTypes": [
                        "ActionRequest",
                        "Chat",
                        "Context",
                        "InternalSearchQuery",
                        "InternalSearchResult",
                        "Disengaged",
                        "InternalLoaderMessage",
                        "Progress",
                        "RenderCardRequest",
                        "AdsQuery",
                        "SemanticSerp",
                        "GenerateContentQuery",
                        "SearchQuery",
                    ],
                    "sliceIds": [
                        "0311tccs0",
                        "gldidentitycf",
                        "qnav2table1",
                        "fltlatest",
                        "ntbkwco",
                        "cntralign2",
                        "cntralign",
                        "bcece403t",
                        "crtcachenock",
                        "romiccf",
                        "rwt2",
                        "cleanjsonctrl",
                        "advtoknmic",
                        "cacblvsdscf",
                        "408mems0",
                        "shopgptctrl",
                        "0329mupcsts0",
                        "unionfdbk",
                        "fpallsticy",
                        "0404redhoo",
                        "saisgrds0",
                        "duptelfix",
                        "fixertel",
                        "ntbkgoldcf",
                        "kcclickthrucf",
                        "cacfrwebt2",
                        "sswebtop1cf",
                        "sswebtop2",
                        "sstopcf"
                    ],
                    "verbosity": "verbose",
                    "scenario": "SERP",
                    "plugins": plugin_params,
                    "traceId": get_ran_hex(32),
                    "gptId": persona.name,
                    "isStartOfSession": self.invocation_id == 3,
                    "message": message,
                    "tone": conversation_style.name.capitalize(),  # Make first letter uppercase
                    "requestId": message_id,
                    "conversationSignature": self.conversation_signature,
                    "participant": {
                        "id": self.client_id,
                    },
                    "conversationId": self.conversation_id,
                },
            ],
            "invocationId": str(self.invocation_id),
            "target": "chat",
            "type": 4,
        }

        if webpage_context:
            self.struct["arguments"][0]["previousMessages"] = [
                {
                    "author": "user",
                    "description": webpage_context,
                    "contextType": "WebPage",
                    "messageType": "Context",
                    "messageId": "discover-web--page-ping-mriduna-----",
                },
            ]
        self.invocation_id += 1
        # print(json.dumps(self.struct, indent=2, ensure_ascii=False))



import json
import os
from typing import List
from typing import Union

from curl_cffi import requests
from curl_cffi.requests import AsyncSession

from .constants import HEADERS_INIT_CONVER
from .exceptions import NotAllowedToAccess


class Conversation:
    def __init__(
        self,
        proxy: Union[str, None] = None,
        async_mode: bool = False,
        cookies: Union[List[dict], None] = None,
    ) -> None:
        if async_mode:
            return
        self.struct: dict = {
            "conversationId": None,
            "clientId": None,
            "conversationSignature": None,
            "secAccessToken": None,
            "result": {"value": "Success", "message": None},
        }
        self.proxy = proxy
        self.session = requests.Session(
            proxy=proxy,
            timeout=900,
            headers=HEADERS_INIT_CONVER,
            impersonate="chrome116"
        )
        if cookies:
            for cookie in cookies:
                self.session.cookies.set(cookie["name"], cookie["value"])
        # Send GET request
        response = self.session.get(
            url=os.environ.get("BING_PROXY_URL")
            or "https://edgeservices.bing.com/edgesvc/turing/conversation/create",
        )
        if response.status_code != 200:
            print(f"Status code: {response.status_code}")
            print(response.text)
            print(response.url)
            raise Exception("Authentication failed")
        try:
            self.struct = response.json()
        except (json.decoder.JSONDecodeError, NotAllowedToAccess) as exc:
            raise Exception(
                "Authentication failed. You have not been accepted into the beta.",
            ) from exc
        if self.struct["result"]["value"] == "UnauthorizedRequest":
            raise NotAllowedToAccess(self.struct["result"]["message"])
        if 'X-Sydney-Encryptedconversationsignature' in response.headers:
            self.struct['secAccessToken'] = response.headers['X-Sydney-Encryptedconversationsignature']

    @staticmethod
    async def create(
        proxy: Union[str, None] = None,
        cookies: Union[List[dict], None] = None,
        conversation_id: Union[str, None] = None,
    ) -> "Conversation":
        self = Conversation(async_mode=True)
        self.struct = {
            "conversationId": None,
            "clientId": None,
            "conversationSignature": None,
            "secAccessToken": None,
            "result": {"value": "Success", "message": None},
        }
        self.proxy = proxy
        proxy = (
            proxy
            or os.environ.get("all_proxy")
            or os.environ.get("ALL_PROXY")
            or os.environ.get("https_proxy")
            or os.environ.get("HTTPS_PROXY")
            or None
        )
        if proxy is not None and proxy.startswith("socks5h://"):
            proxy = "socks5://" + proxy[len("socks5h://") :]

        async with AsyncSession(
            headers=HEADERS_INIT_CONVER,
            proxy=proxy,
            impersonate="chrome116",
        ) as s:
            for cookie in cookies:
                s.cookies.set(cookie["name"], cookie["value"])
            url = os.environ.get("BING_PROXY_URL") or "https://www.bing.com/turing/conversation/create"
            if conversation_id:
                url += f"?conversationId={conversation_id}"

            response = await s.get(url=url)
            if response.status_code != 200:
                print(f"Status code: {response.status_code}")
                print(response.text)
                print(response.url)
                raise Exception("Authentication failed")

            try:
                self.struct = response.json()
            except (json.decoder.JSONDecodeError, NotAllowedToAccess) as exc:
                print(response.text)
                raise Exception(
                    "Authentication failed. You have not been accepted into the beta.",
                ) from exc
            if self.struct["result"]["value"] == "UnauthorizedRequest":
                raise NotAllowedToAccess(self.struct["result"]["message"])
            if 'X-Sydney-Encryptedconversationsignature' in response.headers:
                self.struct['secAccessToken'] = response.headers['X-Sydney-Encryptedconversationsignature']

            return self
